## -*-Pyth-*-
 # ########################################################################
 # FiPy - a finite volume PDE solver in Python
 # 
 # FILE: "pseudoRKQSStepper.py"
 #
 # Author: Jonathan Guyer <guyer@nist.gov>
 # Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 # Author: James Warren   <jwarren@nist.gov>
 #   mail: NIST
 #    www: <http://www.ctcms.nist.gov/fipy/>
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are 
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 # 
 # ########################################################################
 ##

from fipy.steppers.stepper import Stepper

__all__ = ["PseudoRKQSStepper"]

class PseudoRKQSStepper(Stepper):
    """
    Adaptive stepper based on the ``rkqs`` (Runge-Kutta
    "quality-controlled" stepper) algorithm of numerixal Recipes in C: 2nd
    Edition, Section 16.2.
    
    Not really appropriate, since we're not doing Runge-Kutta steps
    in the first place, but works OK.
    """
    def __init__(self, vardata=(), safety=0.9, pgrow=-0.2, pshrink=-0.25, errcon=1.89e-4):
        Stepper.__init__(self, vardata=vardata)
        self.safety = safety
        self.pgrow = pgrow
        self.pshrink = pshrink
        self.errcon = errcon
        
    def _step(self, dt, dtPrev, sweepFn, failFn, *args, **kwargs):
        residual = 1e100
        while residual > 1.:
            residual = sweepFn(vardata=self.vardata, dt=dt, *args, **kwargs)
            
            if residual > 1.:
                # step failed
                failFn(vardata=self.vardata, dt=dt, *args, **kwargs)
                    
                # revert
                for var, eqn, bcs in self.vardata:
                    var.setValue(var.old)
                    
                    dt = max(self.safety * dt * residual**self.pgrow, 0.1 * dt)
                    
                dt = self._lowerBound(dt)

        if residual > self.errcon:
            dtNext = dt * self.safety * residual**self.pshrink
        else:
            dtNext = 5 * dt
            
        return dt, dtNext




__docformat__ = 'restructuredtext'

__all__ = []

try:
    from fipy.viewers.gistViewer import *
    __all__.extend(gistViewer.__all__)
except:
    pass

try:
    from fipy.viewers.gnuplotViewer import *
    __all__.extend(gnuplotViewer.__all__)
except:
    pass

try:
    from fipy.viewers.matplotlibViewer import *
    __all__.extend(matplotlibViewer.__all__)
except:
    pass

try:
    from fipy.viewers.mayaviViewer import *
    __all__.extend(mayaviViewer.__all__)
except:
    pass

from fipy.viewers.multiViewer import *
from fipy.viewers.tsvViewer import *
from fipy.viewers.vtkViewer import *

__all__.extend(multiViewer.__all__)
__all__.extend(tsvViewer.__all__)
__all__.extend(vtkViewer.__all__)

# what about vector variables?

class MeshDimensionError(IndexError):
    pass
    
from fipy.viewers.viewer import AbstractViewer
class DummyViewer(AbstractViewer):
    def plot(self, filename=None):
        pass

def Viewer(vars, title=None, limits={}, FIPY_VIEWER=None, **kwlimits):
    r"""Generic function for creating a `Viewer`. 
    
    The `Viewer` factory will search the module tree and return an instance of
    the first `Viewer` it finds that supports the dimensions of `vars`. Setting
    the '`FIPY_VIEWER`' environment variable to either '`gist`', '`gnuplot`',
    '`matplotlib`', '`tsv`', or '`vtk`' will specify the viewer.
       
    The `kwlimits` or `limits` parameters can be used to constrain the view. For example::
            
        Viewer(vars=some1Dvar, xmin=0.5, xmax=None, datamax=3)
               
    or::
        
        Viewer(vars=some1Dvar, 
               limits={'xmin': 0.5, 'xmax': None, 'datamax': 3})
        
    will return a viewer that displays a line plot from an `x` value
    of 0.5 up to the largest `x` value in the dataset. The data values
    will be truncated at an upper value of 3, but will have no lower
    limit.

    :Parameters:
      vars
        a `CellVariable` or tuple of `CellVariable` objects to plot
      title
        displayed at the top of the `Viewer` window
      limits : dict
        a (deprecated) alternative to limit keyword arguments
      FIPY_VIEWER
        a specific viewer to attempt (possibly multiple times for multiple variables)
      xmin, xmax, ymin, ymax, zmin, zmax, datamin, datamax
        displayed range of data. A 1D `Viewer` will only use `xmin` and
        `xmax`, a 2D viewer will also use `ymin` and `ymax`, and so on. All
        viewers will use `datamin` and `datamax`. Any limit set to a
        (default) value of `None` will autoscale.
      
    """
    import os

    if type(vars) not in [type([]), type(())]:
        vars = [vars]
    vars = list(vars)
    
    if FIPY_VIEWER is None and 'FIPY_VIEWER' in os.environ:
        FIPY_VIEWER = os.environ['FIPY_VIEWER']

    if FIPY_VIEWER == "dummy":
        return DummyViewer(vars=vars)

    errors = []

    attempts = []
    viewers = []
    
    emptyvars = [var for var in vars if var.mesh.numberOfCells == 0]
    vars = [var for var in vars if var.mesh.numberOfCells > 0]
    
    if len(emptyvars):
        viewers.append(DummyViewer(vars=emptyvars))
    
    import pkg_resources
    for ep in pkg_resources.iter_entry_points(group='fipy.viewers', 
                                              name=FIPY_VIEWER):
                                                  
        attempts.append(ep.name)
        
        try:
            ViewerClass = ep.load()
            
            while len(vars) > 0:
                viewer = ViewerClass(vars=vars, title=title, limits=limits, **kwlimits)
                
                for var in viewer.vars:
                    vars.remove(var)
                
                viewers.append(viewer)
                
            break
        except Exception as s:
            errors.append("%s: %s" % (ep.name, s))

    if len(attempts) == 0:
        if FIPY_VIEWER is not None:
            raise ImportError("`%s` viewer not found" % FIPY_VIEWER)
        else:
            raise ImportError("No viewers found. Run `python setup.py egg_info` or similar.")
    
    if len(vars) > 0:
        raise ImportError("Failed to import a viewer: %s" % str(errors))
            
    if len(viewers) > 1:
        return MultiViewer(viewers = viewers)
    else:
        return viewers[0]
        
__all__.extend(["MeshDimensionError", "DummyViewer", "Viewer"])



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "faceVariable.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution
 #  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

from fipy.variables.meshVariable import _MeshVariable
from fipy.tools import numerix
from fipy.tools.decorators import getsetDeprecated

__all__ = ["FaceVariable"]

class FaceVariable(_MeshVariable):
    @property
    def _variableClass(self):
        return FaceVariable

    def _getShapeFromMesh(mesh):
        """
        Return the shape of this variable type, given a particular mesh.
        """
        return (mesh.numberOfFaces,)
    _getShapeFromMesh = staticmethod(_getShapeFromMesh)
        
    def _getArithmeticBaseClass(self, other = None):
        """
        Given `self` and `other`, return the desired base class for an operation
        result.
        """
        if other is None:
            return FaceVariable
            
        return _MeshVariable._getArithmeticBaseClass(self, other)

    def copy(self):
        return self._getArithmeticBaseClass()(mesh=self.mesh, 
                                              name=self.name + "_copy", 
                                              value=self.value)

    @property
    def globalValue(self):
        return self._getGlobalValue(self.mesh._localNonOverlappingFaceIDs, 
                                    self.mesh._globalNonOverlappingFaceIDs)

    def setValue(self, value, unit = None, where = None):
        _MeshVariable.setValue(self, value=self._globalToLocalValue(value), unit=unit, where=where)

    @getsetDeprecated
    def getDivergence(self):
        return self.divergence

    @property
    def divergence(self):
        """
            >>> from fipy.meshes import Grid2D
            >>> from fipy.variables.cellVariable import CellVariable
            >>> mesh = Grid2D(nx=3, ny=2)
            >>> var = CellVariable(mesh=mesh, value=range(3*2))
            >>> print var.faceGrad.divergence
            [ 4.  3.  2. -2. -3. -4.]
            
        """
        if not hasattr(self, '_divergence'):
            from fipy.variables.addOverFacesVariable import _AddOverFacesVariable

            s = (slice(0,None,None),) + (numerix.newaxis,) * (len(self.shape) - 2) + (slice(0,None,None),)
            self._divergence = _AddOverFacesVariable((self * self.mesh._orientedAreaProjections[s]).sum(0))

        return self._divergence

    @property
    def _globalNumberOfElements(self):
        return self.mesh.globalNumberOfFaces
        
    @property
    def _globalOverlappingIDs(self):
        return self.mesh._globalOverlappingFaceIDs
        
    @property
    def _localNonOverlappingIDs(self):
        return self.mesh._localNonOverlappingFaceIDs

def _test(): 
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()
    
if __name__ == "__main__": 
    _test() 




#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - a finite volume PDE solver in Python
 # 
 #  FILE: "baseBinaryTerm.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This document was prepared at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this document is not subject to copyright
 # protection and is in the public domain.  summationTerm.py
 # is an experimental work.  NIST assumes no responsibility whatsoever
 # for its use by other parties, and makes no guarantees, expressed
 # or implied, about its quality, reliability, or any other characteristic.
 # We would appreciate acknowledgement if the document is used.
 # 
 # This document can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

__all__ = []

import os

from fipy.terms.term import Term
from fipy.terms.explicitSourceTerm import _ExplicitSourceTerm
from fipy.terms import ExplicitVariableError

class _AbstractBinaryTerm(Term):
    def __init__(self, term, other):

        if not isinstance(other, Term):
            other = _ExplicitSourceTerm(coeff=other, var=term.var)

        self.term = term
        self.other = other

        if term.var is None:
            if other.var is None:
                pass
            else:
                raise ExplicitVariableError
        else:
            if other.var is None:
                if isinstance(other, _ExplicitSourceTerm):
                    other.var = term.var
                else:
                    raise ExplicitVariableError

        Term.__init__(self, var=self._vars[0])
    
    def _addNone(self, arg0, arg1):
        if arg0 is None and arg1 is None:
            return None
        elif arg0 is None:
            return arg1
        elif arg1 is None:
            return arg0
        else:
            return arg0 + arg1

    def __neg__(self):
        r"""
         Negate a `_BinaryTerm`.

           >>> -(__NonDiffusionTerm(coeff=1.) - __NonDiffusionTerm(coeff=2.))
           (__NonDiffusionTerm(coeff=-1.0) + __NonDiffusionTerm(coeff=2.0))

        """

        return (-self.term) + (-self.other)

    def _calcVars(self):
        return list(set(self.term._vars + self.other._vars))

    @property
    def _vars(self):
        if not hasattr(self, '_internalVars'):
            self._internalVars = self._calcVars()
        return self._internalVars

    @property
    def _transientVars(self):
        return self.term._transientVars + self.other._transientVars

    @property
    def _diffusionVars(self):
        return self.term._diffusionVars + self.other._diffusionVars

    def _checkVar(self, var):
        self.term._checkVar(var)
        self.other._checkVar(var)
    

from fipy.terms.nonDiffusionTerm import _NonDiffusionTerm
class __NonDiffusionTerm(_NonDiffusionTerm):
    """
    Dummy subclass for tests
    """
    pass 

def _test(): 
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



#!/usr/bin/env python

## 
 # -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #  Author: James O'Beirne <james.obeirne@gmail.com>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution
 #  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

__all__ = []

from fipy.tools.dimensions.physicalField import PhysicalField
from fipy.tools import numerix

class _UniformOrigin(object):
    """
    Used to calculate the origin for uniform grids in a
    dimensionally-independent way.
    """

    @staticmethod
    def calcOrigin(origin, offset, ds, scale):
        newOrigin  = PhysicalField(value=origin)
        newOrigin /= scale

        if type(offset) in [int, float]:
            newOrigin += offset * ds[0]
        else:
            newOrigin += [[o*float(d)] for o, d in zip(offset, ds)]

        return newOrigin

class _DOffsets(object):
    """
    For use by non-uniform grid builders.
    """

    @staticmethod
    def calcDOffsets(ds, ns, offset):
        """
        :Parameters:
            - `ds`: A list, e.g. [dx, dy]
            - `ns`: A list, e.g. [nx, ny, nz]
            - `offset`

        :Returns:
            - `offsetList`: a list which contains the analogous scalars to
              `XOffset`, `YOffset`, and `ZOffset`, whichever are applicable for
              the dimensionality.
            - `newDs`: a list containing proper [dx, [dy, ...]] values
        """
        offsetList = []
        newDs = []

        if type(offset) in [int, float]:
            offset = [offset]

        for d, n, i in zip(ds, ns, list(range(len(ds)))):
            if numerix.getShape(d) is not ():
                offsetList.append(numerix.sum(d[0:offset[i]]))
                newDs.append(d[offset[i]:offset[i] + n])
            else:
                if len(offset) == 1:
                    offsetList.append(d * offset[0])
                else:
                    offsetList.append(d * offset[i])

                newDs.append(d)

        return offsetList, newDs

class _AbstractNumPts(object):
    """
    Interface definition for NumPtsCalculators.
    """

    @staticmethod
    def calcNs(ns, ds):
        raise NotImplementedError
 
class _NonuniformNumPts(_AbstractNumPts):
    """
    For use by non-uniform grid builders.
    """

    @staticmethod
    def calcNs(ns, ds):
        axis = ["x", "y", "z"][:len(ns)]
        newNs = []

        for a, d, n in zip(axis, ds, ns):
            newNs.append(_NonuniformNumPts._calcNumPts(d=d, n=n, axis=a))

        return newNs

    @staticmethod
    def _calcNumPts(d, n = None, axis = "x"):
        """
        Calculate the number of cells along the specified axis, based
        on either the specified number or on the number elements in the
        cell  `d` spacings.
        
        Used by the `Grid` meshes.

        This tests a bug that was occuring with PeriodicGrid1D when
        using a numpy float as the argument for the grid spacing.

           >>> from fipy.meshes.periodicGrid1D import PeriodicGrid1D
           >>> PeriodicGrid1D(nx=2, dx=numerix.float32(1.))
           PeriodicGrid1D(dx=1.0, nx=2)
        """

        if type(d) in [int, float] or numerix.shape(d) == ():
            n = int(n or 1)
        else:
            n = int(n or len(d))
            if n != len(d) and len(d) != 1:
                raise IndexError("n%s != len(d%s)" % (axis, axis))
                
        return n

class _UniformNumPts(_AbstractNumPts):
    """
    For use by uniform grid builders.
    """
 
    @staticmethod
    def calcNs(ns, ds):
        return [int(x) for x in ns]

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "fixedFlux.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #  Author: James Warren <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

from fipy.tools import numerix

from fipy.boundaryConditions.boundaryCondition import BoundaryCondition
from fipy.boundaryConditions.fixedValue import FixedValue
from fipy.tools import vector

__all__ = ["FixedFlux"]

class FixedFlux(BoundaryCondition):
    r"""

    The `FixedFlux` boundary condition adds a contribution, equivalent to a
    fixed flux (Neumann condition), to the equation's RHS vector.  The
    contribution, given by `value`, is only added to entries corresponding to
    the specified `faces`, and is weighted by the face areas.
       
    """
    
    def __init__(self,faces,value):
        """
        Creates a `FixedFlux` object.
        
        :Parameters:
            - `faces`: A `list` or `tuple` of `Face` objects to which this condition applies.
            - `value`: The value to impose.
            
        """
        BoundaryCondition.__init__(self,faces,value)
        ## The extra index [self.faces.value] makes self.contribution the same length as self.adjacentCellIDs
        self.contribution = (self.value * self.faces.mesh._faceAreas)[self.faces.value]
        
    def _buildMatrix(self, SparseMatrix, Ncells, MaxFaces, coeff):
        """Leave **L** unchanged and add gradient to **b**
        
        :Parameters:
          - `SparseMatrix`: *unused* (empty matrix)
          - `Ncells`:       Size of **b**-vector
          - `MaxFaces`:     *unused*
          - `coeff`:        *unused*
        """
        
        bb = numerix.zeros((Ncells,),'d')

        if not self.boundaryConditionApplied:
            vector.putAdd(bb, self.adjacentCellIDs, -self.contribution)
            self.boundaryConditionApplied = True
         
        return (0, bb)

    def _getDerivative(self, order):
        if order == 1:
            return FixedValue(self.faces, self.value)
        else:
            return BoundaryCondition._getDerivative(self, order)

    def _test(self):
        """
        The following tests check that self.contributions is the same length as
        self.adjacentCellIDs.

           >>> from fipy import *
           >>> m = Grid1D(nx = 10)
           >>> v = FaceVariable(mesh=m)
           >>> bc = FixedFlux(value=v.globalValue[-1], faces=m.facesRight)
           >>> len(bc.contribution) == len(bc.adjacentCellIDs)
           True
           >>> bc = FixedFlux(value=v, faces=m.facesRight)
           >>> len(bc.contribution) == len(bc.adjacentCellIDs)
           True
        """

def _test(): 
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()
    
if __name__ == "__main__": 
    _test()



from fipy.variables.variable import *
from fipy.variables.cellVariable import *
from fipy.variables.faceVariable import *
from fipy.variables.scharfetterGummelFaceVariable import *
from fipy.variables.modularVariable import *
from fipy.variables.betaNoiseVariable import *
from fipy.variables.exponentialNoiseVariable import *
from fipy.variables.gammaNoiseVariable import *
from fipy.variables.gaussianNoiseVariable import *
from fipy.variables.uniformNoiseVariable import *
from fipy.variables.histogramVariable import *
from fipy.variables.surfactantVariable import *
from fipy.variables.surfactantConvectionVariable import *
from fipy.variables.distanceVariable import *

__all__ = []
__all__.extend(variable.__all__)
__all__.extend(cellVariable.__all__)
__all__.extend(faceVariable.__all__)
__all__.extend(scharfetterGummelFaceVariable.__all__)
__all__.extend(modularVariable.__all__)
__all__.extend(betaNoiseVariable.__all__)
__all__.extend(exponentialNoiseVariable.__all__)
__all__.extend(gammaNoiseVariable.__all__)
__all__.extend(gaussianNoiseVariable.__all__)
__all__.extend(uniformNoiseVariable.__all__)
__all__.extend(histogramVariable.__all__)
__all__.extend(surfactantVariable.__all__)
__all__.extend(surfactantConvectionVariable.__all__)
__all__.extend(distanceVariable.__all__)



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "trilinosAztecOOSolver.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #  Author: Maxsim Gibiansky <maxsim.gibiansky@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

import os

from PyTrilinos import AztecOO

from fipy.solvers.trilinos.trilinosSolver import TrilinosSolver
from fipy.solvers.trilinos.preconditioners.jacobiPreconditioner import JacobiPreconditioner

__all__ = ["TrilinosAztecOOSolver"]

class TrilinosAztecOOSolver(TrilinosSolver):

    """
    .. attention:: This class is abstract, always create on of its subclasses. It provides the code to call all solvers from the Trilinos AztecOO package.

    """
      
    def __init__(self, tolerance=1e-10, iterations=1000, precon=JacobiPreconditioner()):
        """
        :Parameters:
          - `tolerance`: The required error tolerance.
          - `iterations`: The maximum number of iterative steps to perform.
          - `precon`: Preconditioner object to use. 

        """
        if self.__class__ is TrilinosAztecOOSolver:
            raise NotImplementedError("can't instantiate abstract base class")
            
        TrilinosSolver.__init__(self, tolerance=tolerance,
                                iterations=iterations, precon=None)
        self.preconditioner = precon

    def _solve_(self, L, x, b):

        Solver = AztecOO.AztecOO(L, x, b)
        Solver.SetAztecOption(AztecOO.AZ_solver, self.solver)

##        Solver.SetAztecOption(AztecOO.AZ_kspace, 30)

        Solver.SetAztecOption(AztecOO.AZ_output, AztecOO.AZ_none)

        if self.preconditioner is not None:
            self.preconditioner._applyToSolver(solver=Solver, matrix=L)
        else:
            Solver.SetAztecOption(AztecOO.AZ_precond, AztecOO.AZ_none)

        output = Solver.Iterate(self.iterations, self.tolerance)

        if self.preconditioner is not None:
            if hasattr(self.preconditioner, 'Prec'):
                del self.preconditioner.Prec

        if 'FIPY_VERBOSE_SOLVER' in os.environ:
            status = Solver.GetAztecStatus()

            from fipy.tools.debug import PRINT        
            PRINT('iterations: %d / %d' % (status[AztecOO.AZ_its], self.iterations))
            failure = {AztecOO.AZ_normal : 'AztecOO.AZ_normal',
                       AztecOO.AZ_param : 'AztecOO.AZ_param',
                       AztecOO.AZ_breakdown : 'AztecOO.AZ_breakdown',
                       AztecOO.AZ_loss : 'AztecOO.AZ_loss',
                       AztecOO.AZ_ill_cond : 'AztecOO.AZ_ill_cond',
                       AztecOO.AZ_maxits : 'AztecOO.AZ_maxits'}

            PRINT('failure',failure[status[AztecOO.AZ_why]])
                               
            PRINT('AztecOO.AZ_r:',status[AztecOO.AZ_r])
            PRINT('AztecOO.AZ_scaled_r:',status[AztecOO.AZ_scaled_r])
            PRINT('AztecOO.AZ_rec_r:',status[AztecOO.AZ_rec_r])
            PRINT('AztecOO.AZ_solve_time:',status[AztecOO.AZ_solve_time])
            PRINT('AztecOO.AZ_Aztec_version:',status[AztecOO.AZ_Aztec_version])
        
        return output



## -*-Pyth-*-
 # #############################################################################
 # FiPy - a finite volume PDE solver in Python
 # 
 # FILE: "operatorVariable.py"
 #
 # Author: Jonathan Guyer <guyer@nist.gov>
 # Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 # Author: James Warren   <jwarren@nist.gov>
 #   mail: NIST
 #    www: <http://www.ctcms.nist.gov/fipy/>
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are 
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 # 
 # #############################################################################
 ##

__docformat__ = 'restructuredtext'

__all__ = []

import sys

from fipy.variables.variable import Variable
from fipy.tools import numerix
from fipy.tools.decorators import getsetDeprecated

def _OperatorVariableClass(baseClass=object):
    class _OperatorVariable(baseClass):
        def __init__(self, op, var, opShape=(), canInline=True, unit=None, inlineComment=None, *args, **kwargs):
            self.op = op
            self.var = var
            self.opShape = opShape
            self._unit = unit
            self.canInline = canInline  #allows for certain functions to opt out of --inline
            baseClass.__init__(self, value=None, *args, **kwargs)
            self.name = ''
            for var in self.var:    #C does not accept units
                if not var.unit.isDimensionless():
                    self.canInline = False
                    break

            for aVar in self.var:
                self._requires(aVar)
            
            self.dontCacheMe()

            self.comment = inlineComment

        def __setitem__(self, index, value):
            raise TypeError("The value of an `_OperatorVariable` cannot be assigned")
            
        def setValue(self, value, unit=None, where=None):
            raise TypeError("The value of an `_OperatorVariable` cannot be assigned")
        
        def _calcValue(self):
            if not self.canInline:
                return self._calcValue_()
            else:
                from fipy.tools import inline
                if inline.doInline:
                    return self._execInline(comment=self.comment)
                else:
                    return self._calcValue_()

        def _calcValue_(self):
            pass

        def _isCached(self):
            return (Variable._isCached(self) 
                    or (len(self.subscribedVariables) > 1 and not self._cacheNever))

        def _getCstring(self, argDict={}, id="", freshen=False):
            if self.canInline: # and not self._isCached():
                s = self._getRepresentation(style="C", argDict=argDict, id=id, freshen=freshen)
            else:
                s = baseClass._getCstring(self, argDict=argDict, id=id)
            if freshen:
                self._markFresh()
              
            return s
                
        def _getRepresentation(self, style="__repr__", argDict={}, id=id, freshen=False):
            """

            :Parameters:
                
              - `style`: one of `'__repr__'`, `'name'`, `'TeX'`, `'C'`

            """
            import opcode
            
            def __var(i):
                v = self.var[i]
                if style == "__repr__":
                    result = repr(v)
                elif style == "name":
                    if isinstance(v, Variable):
                        result = v.name
                        if len(result) == 0:
                            # The string form of a variable
                            # would probably be too long and messy.
                            # Just give shorthand.
                            result = "%s(...)" % v.__class__.__name__
                    elif type(v) in (type(1), type(1.)):
                        result = repr(v)
                    else:
                        # The string form of anything but a
                        # number would be too long and messy.
                        # Just give shorthand.
                        result = "<...>"

                elif style == "TeX":
                    raise Exception("TeX style not yet implemented")
                elif style == "C":
                    if not v._isCached():
                        result = v._getCstring(argDict, id=id + str(i), freshen=freshen)
                        if isinstance(v, Variable):
                            v._value = None
                        else:
                            v.value = None
                    else:
                        result = v._variableClass._getCstring(v, argDict,
                                                                   id=id + str(i),
                                                                   freshen=False)
                else:
                    raise SyntaxError("Unknown style: %s" % style)
                    
                return result

            if isinstance(self.op, numerix.ufunc):
                return "%s(%s)" % (self.op.__name__, ", ".join([__var(i) for i in range(len(self.var))]))
            
            if sys.version_info < (3,0):
                bytecodes = [ord(byte) for byte in self.op.__code__.co_code]
            else:
                bytecodes = list(self.op.__code__.co_code)
                
            def _popIndex():
                return bytecodes.pop(0) + bytecodes.pop(0) * 256
            
            stack = []
                
            unop = {
                10: "+", 11: "-", 12: "not ", 15: "~"
            }
            
            binop = {
                19: "**", 20: "*", 21: "/", 22: "%", 23: "+", 24: "-", 26: "//", 27: "/",
                        62: "<<", 63: ">>", 64: "&", 65: "^", 66: "|", 106: "=="
            }
            
            while len(bytecodes) > 0:
                bytecode = bytecodes.pop(0)
                if opcode.opname[bytecode] == 'UNARY_CONVERT':
                    stack.append("`" + stack.pop() + "`")
                elif opcode.opname[bytecode] == 'BINARY_SUBSCR':
                    stack.append(stack.pop(-2) + "[" + stack.pop() + "]")
                elif opcode.opname[bytecode] == 'RETURN_VALUE':
                    s = stack.pop()
                    if style == 'C':
                        return s.replace('numerix.', '').replace('arc', 'a')
                    else:
                        return s
                elif opcode.opname[bytecode] == 'LOAD_CONST':
                    stack.append(self.op.__code__.co_consts[_popIndex()])
                elif opcode.opname[bytecode] == 'LOAD_ATTR':
                    stack.append(stack.pop() + "." + self.op.__code__.co_names[_popIndex()])
                elif opcode.opname[bytecode] == 'COMPARE_OP':
                    stack.append(stack.pop(-2) + " " + opcode.cmp_op[_popIndex()] + " " + stack.pop())
                elif opcode.opname[bytecode] == 'LOAD_GLOBAL':
                    counter = _popIndex()
                    stack.append(self.op.__code__.co_names[counter])
                elif opcode.opname[bytecode] == 'LOAD_FAST':
                    stack.append(__var(_popIndex()))
                elif opcode.opname[bytecode] == 'CALL_FUNCTION':    
                    args = []
                    for j in range(bytecodes.pop(1)):
                        # keyword parameters
                        args.insert(0, stack.pop(-2) + " = " + stack.pop())
                    for j in range(bytecodes.pop(0)):
                        # positional parameters
                        args.insert(0, stack.pop())
                    stack.append(stack.pop() + "(" + ", ".join(args) + ")")
                elif opcode.opname[bytecode] == 'LOAD_DEREF':
                    free = self.op.__code__.co_cellvars + self.op.__code__.co_freevars
                    stack.append(free[_popIndex()])
                elif bytecode in unop:
                    stack.append(unop[bytecode] + '(' + stack.pop() + ')')
                elif bytecode in binop:
                    stack.append(stack.pop(-2) + " " + binop[bytecode] + " " + stack.pop())
                else:
                    raise SyntaxError("Unknown bytecode: %s in %s: %s" % (
                       repr(bytecode), 
                       repr([ord(byte) for byte in self.op.__code__.co_code]),
                       "FIXME"))
                
        def __repr__(self):
            return self._getRepresentation()
            
        def __reduce__(self):
            """
            Allows _OperatorVariables to be pickled
            """
            state =  self.__getstate__()
            if 'mesh' in list(state.keys()):
                args = (state['mesh'],)
            else:
                args = ()
                        
            return (self._variableClass, args, self.__getstate__())

        @getsetDeprecated
        def getName(self):
            return self.name

        def _getName(self):
            name = baseClass._getName(self)
            if len(name) == 0:
                name = self._getRepresentation(style="name")
            return name
   
        name = property(_getName, baseClass._setName)

        @property
        def shape(self):
            if self.opShape is not None:
                return self.opShape
            else:
                return baseClass.getShape(self)
##             return baseClass.getShape(self) or self.opShape

    return _OperatorVariable
    
def _testBinOp(self):
    """
    Test of _getRepresentation
    
        >>> v1 = Variable((1,2,3,4))
        >>> v2 = Variable((5,6,7,8))
        >>> v3 = Variable((9,10,11,12))
        >>> v4 = Variable((13,14,15,16))

        >>> (v1 * v2)._getRepresentation()
        '(Variable(value=array([1, 2, 3, 4])) * Variable(value=array([5, 6, 7, 8])))'
        
        >>> (v1 * v2)._getRepresentation(style='C', id="")
        '(var0[i] * var1[i])'
        
        >>> (v1 * v2 + v3 * v4)._getRepresentation(style='C', id="")
        '((var00[i] * var01[i]) + (var10[i] * var11[i]))'
        
        >>> (v1 - v2)._getRepresentation(style='C', id="")
        '(var0[i] - var1[i])'

        >>> (v1 / v2)._getRepresentation(style='C', id="")
        '(var0[i] / var1[i])'

        >>> (v1 - 1)._getRepresentation(style='C', id="")
        '(var0[i] - var1)'
            
        >>> (5 * v2)._getRepresentation(style='C', id="")
        '(var0[i] * var1)'

        >>> (v1 / v2 - v3 * v4 + v1 * v4)._getRepresentation(style='C', id="")
        '(((var000[i] / var001[i]) - (var010[i] * var011[i])) + (var10[i] * var11[i]))'
        
    Check that unit works for a binOp

        >>> (Variable(value="1 m") * Variable(value="1 s")).unit
        <PhysicalUnit s*m>

        >>> (Variable(value="1 m") / Variable(value="0 s")).unit
        <PhysicalUnit m/s>

        >>> a = -(numerix.sin(Variable() * Variable()))

    Check that getTypeCode() works as expected.

        >>> a = Variable(1.) * Variable(1)
        >>> print a.getsctype() == numerix.float64
        1

    The following test is to correct an `--inline` bug that was
    being thrown by the Cahn-Hilliard example. The fix for this
    bug was to add the last line to the following code in
    `_getRepresentation()`.
    
        >>> ##elif style == "C":
        >>> ##    counter = _popIndex()
        >>> ##    if not self.var[counter]._isCached():
        >>> ##        stack.append(self.var[counter]._getCstring(argDict, id=id + str(counter), freshen=freshen))
        >>> ##        self.var[counter].value=None

    This is the test that fails if the last line above is removed
    from `_getRepresentation()`, the `binOp.value` statement
    below will return `1.0` and not `0.5`.
        
        >>> from fipy import numerix
        >>> def doBCs(binOp):
        ...     unOp1 = -binOp
        ...     print binOp.value
        >>> var = Variable(1.)
        >>> binOp = 1. * var
        >>> unOp = -binOp
        >>> print unOp.value
        -1.0
        >>> doBCs(binOp)
        1.0
        >>> var.value = (0.5)
        >>> print unOp.value
        -0.5
        >>> unOp2 = -binOp
        >>> print binOp.value
        0.5

        >>> from fipy.variables.cellVariable import CellVariable
        >>> from fipy.variables.faceVariable import FaceVariable
        
        >>> from fipy.meshes import Grid2D
        >>> mesh = Grid2D(nx=3)
        
        
    `CellVariable` * CellVariable
    
        >>> cv = CellVariable(mesh=mesh, value=(0, 1, 2))
        >>> cvXcv = cv * cv
        >>> print cvXcv
        [0 1 4]
        >>> print isinstance(cvXcv, CellVariable)
        1
    
    `CellVariable` * FaceVariable
    
        >>> fv = FaceVariable(mesh=mesh, value=(0, 1, 2, 3, 4, 5, 6, 7, 8, 9))
        >>> fvXcv = fv * cv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> cvXfv = cv * fv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        
    rank-0 `CellVariable` * rank-1 `CellVariable`
    
        >>> vcv = CellVariable(mesh=mesh, value=((0, 1, 2), (1, 2, 3)), rank=1)
        >>> vcvXcv = vcv * cv
        >>> print vcvXcv
        [[0 1 4]
         [0 2 6]]
        >>> print isinstance(vcvXcv, CellVariable)
        1
        >>> print vcvXcv.rank
        1
        >>> cvXvcv = cv * vcv
        >>> print cvXvcv
        [[0 1 4]
         [0 2 6]]
        >>> print isinstance(cvXvcv, CellVariable)
        1
        >>> print cvXvcv.rank
        1

    `CellVariable` * rank-1 `FaceVariable`

        >>> vfv = FaceVariable(mesh=mesh, value=((0,1,2,3,1,2,3,6,2,1),
        ...                                      (1,2,3,4,3,4,5,9,6,3)), rank=1)
        >>> vfvXcv = vfv * cv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> cvXvfv = cv * vfv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

    `CellVariable` * Scalar
    
        >>> cvXs = cv * 3
        >>> print cvXs
        [0 3 6]
        >>> print isinstance(cvXs, CellVariable)
        1
        >>> sXcv = 3 * cv
        >>> print sXcv
        [0 3 6]
        >>> print isinstance(sXcv, CellVariable)
        1

    `CellVariable` * Vector
    
        >>> cvXv2 = cv * (3,2)
        >>> print cvXv2
        [[0 3 6]
         [0 2 4]]
        >>> cvXv2 = cv * [[3], [2]]
        >>> print cvXv2
        [[0 3 6]
         [0 2 4]]
        >>> print isinstance(cvXv2, CellVariable)
        1
        >>> print cvXv2.rank
        1
        >>> v2Xcv = (3,2) * cv
        >>> print v2Xcv
        [[0 3 6]
         [0 2 4]]
        >>> v2Xcv = [[3], [2]] * cv
        >>> print v2Xcv
        [[0 3 6]
         [0 2 4]]
        >>> print isinstance(v2Xcv, CellVariable)
        1
        >>> print v2Xcv.rank
        1
        
        >>> cvXv3 = cv * (3,2,1)
        >>> print cvXv3
        [0 2 2]
        >>> print isinstance(cvXv3, CellVariable)
        1
        >>> v3Xcv = (3,2,1) * cv
        >>> print v3Xcv
        [0 2 2]
        >>> print isinstance(v3Xcv, CellVariable)
        1
        
        >>> cvXv4 = cv * (3,2,1,0) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        TypeError: can't multiply sequence to non-int
        >>> v4Xcv = (3,2,1,0) * cv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        TypeError: can't multiply sequence to non-int


    `CellVariable` * `Variable` Scalar
    
        >>> cvXsv = cv * Variable(value=3)
        >>> print cvXsv
        [0 3 6]
        >>> print isinstance(cvXsv, CellVariable)
        1
        >>> svXcv = Variable(value=3) * cv
        >>> print svXcv
        [0 3 6]
        >>> print isinstance(svXcv, CellVariable)
        1
    
    `binOp` `CellVariable` * `binOp` `Variable` Scalar

        >>> cvcvXsvsv = (cv * cv) * (Variable(value=3) * Variable(value=3))
        >>> print cvcvXsvsv
        [ 0  9 36]
        >>> print isinstance(cvcvXsvsv, CellVariable)
        1
        >>> svsvXcvcv = (Variable(value=3) * Variable(value=3)) * (cv * cv)
        >>> print svsvXcvcv
        [ 0  9 36]
        >>> print isinstance(svsvXcvcv, CellVariable)
        1
        
    `CellVariable` * `Variable` Vector
        
        >>> cvXv2v = cv * Variable(value=(3,2))
        >>> print cvXv2v
        [[0 3 6]
         [0 2 4]]
        >>> cvXv2v = cv * Variable(value=((3,),(2,)))
        >>> print cvXv2v
        [[0 3 6]
         [0 2 4]]
        >>> print isinstance(cvXv2v, CellVariable)
        1
        >>> print cvXv2v.rank
        1
        >>> v2vXcv = Variable(value=(3,2)) * cv
        >>> v2vXcv = Variable(value=((3,),(2,))) * cv
        >>> print v2vXcv
        [[0 3 6]
         [0 2 4]]
        >>> print isinstance(v2vXcv, CellVariable)
        1
        >>> print v2vXcv.rank
        1
        
        >>> cvXv3v = cv * Variable(value=(3,2,1))
        >>> print cvXv3v
        [0 2 2]
        >>> print isinstance(cvXv3v, CellVariable)
        1
        >>> v3vXcv = Variable(value=(3,2,1)) * cv
        >>> print v3vXcv
        [0 2 2]
        >>> print isinstance(v3vXcv, CellVariable)
        1

        >>> cvXv4v = cv * Variable(value=(3,2,1,0)) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v4vXcv = Variable(value=(3,2,1,0)) * cv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        

    `CellVariable` * CellGradVariable
    
        >>> cvXcgv = cv * cv.grad
        >>> print cvXcgv
        [[ 0.  1.  1.]
         [ 0.  0.  0.]]
        >>> print isinstance(cvXcgv, CellVariable)
        1
        >>> print cvXcgv.rank
        1
        
    `FaceVariable` * FaceVariable

        >>> fvXfv = fv * fv
        >>> print fvXfv
        [ 0  1  4  9 16 25 36 49 64 81]
        >>> print isinstance(fvXfv, FaceVariable)
        1

    `FaceVariable` * rank-1 `CellVariable`

        >>> vcvXfv = vcv * fv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> fvXvcv = fv * vcv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

    `FaceVariable` * rank-1 `FaceVariable`

        >>> vfvXfv = vfv * fv
        >>> print vfvXfv
        [[ 0  1  4  9  4 10 18 42 16  9]
         [ 0  2  6 12 12 20 30 63 48 27]]
        >>> print isinstance(vfvXfv, FaceVariable)
        1
        >>> print vfvXfv.rank
        1
        >>> fvXvfv = fv * vfv
        >>> print fvXvfv
        [[ 0  1  4  9  4 10 18 42 16  9]
         [ 0  2  6 12 12 20 30 63 48 27]]
        >>> print isinstance(fvXvfv, FaceVariable)
        1
        >>> print fvXvfv.rank
        1

    `FaceVariable` * Scalar

        >>> fvXs = fv * 3
        >>> print fvXs
        [ 0  3  6  9 12 15 18 21 24 27]
        >>> print isinstance(fvXs, FaceVariable)
        1
        >>> sXfv = 3 * fv
        >>> print sXfv
        [ 0  3  6  9 12 15 18 21 24 27]
        >>> print isinstance(sXfv, FaceVariable)
        1

    `FaceVariable` * Vector

        >>> fvXv2 = fv * (3,2)
        >>> print fvXv2
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> fvXv2 = fv * ((3,),(2,))
        >>> print fvXv2
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> print isinstance(fvXv2, FaceVariable)
        1
        >>> print fvXv2.rank
        1
        >>> v2Xfv = (3,2) * fv
        >>> print v2Xfv
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> v2Xfv = ((3,),(2,)) * fv
        >>> print v2Xfv
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> print isinstance(v2Xfv, FaceVariable)
        1
        >>> print v2Xfv.rank
        1
        
        >>> fvXv3 = fv * (3,2,1) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v3Xfv = (3,2,1) * fv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

        >>> fvXv10 = fv * (9,8,7,6,5,4,3,2,1,0)
        >>> print fvXv10
        [ 0  8 14 18 20 20 18 14  8  0]
        >>> print isinstance(fvXv10, FaceVariable)
        1
        >>> v10Xfv = (9,8,7,6,5,4,3,2,1,0) * fv
        >>> print v10Xfv
        [ 0  8 14 18 20 20 18 14  8  0]
        >>> print isinstance(v10Xfv, FaceVariable)
        1

    `FaceVariable` * `Variable` Scalar

        >>> fvXsv = fv * Variable(value=3)
        >>> print fvXsv
        [ 0  3  6  9 12 15 18 21 24 27]
        >>> print isinstance(fvXsv, FaceVariable)
        1
        >>> svXfv = Variable(value=3) * fv
        >>> print svXfv
        [ 0  3  6  9 12 15 18 21 24 27]
        >>> print isinstance(svXfv, FaceVariable)
        1

    `FaceVariable` * `Variable` Vector
        
        >>> fvXv2v = fv * Variable(value=(3,2))
        >>> print fvXv2v
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> fvXv2v = fv * Variable(value=((3,),(2,)))
        >>> print fvXv2v
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> print isinstance(fvXv2v, FaceVariable)
        1
        >>> print fvXv2v.rank
        1
        >>> v2vXfv = Variable(value=(3,2)) * fv
        >>> print v2vXfv
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> v2vXfv = Variable(value=((3,),(2,))) * fv
        >>> print v2vXfv
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> print isinstance(v2vXfv, FaceVariable)
        1
        >>> print v2vXfv.rank
        1
        
        >>> fvXv3v = fv * Variable(value=(3,2,1)) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v3vXfv = Variable(value=(3,2,1)) * fv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

        >>> fvXv10v = fv * Variable(value=(9,8,7,6,5,4,3,2,1,0))
        >>> print fvXv10v
        [ 0  8 14 18 20 20 18 14  8  0]
        >>> print isinstance(fvXv10v, FaceVariable)
        1
        >>> v10vXfv = Variable(value=(9,8,7,6,5,4,3,2,1,0)) * fv
        >>> print v10vXfv
        [ 0  8 14 18 20 20 18 14  8  0]
        >>> print isinstance(v10vXfv, FaceVariable)
        1

        
        
    rank-1 `CellVariable` * rank-1 `CellVariable`

        >>> vcvXvcv = vcv * vcv
        >>> print vcvXvcv
        [[0 1 4]
         [1 4 9]]
        >>> print isinstance(vcvXvcv, CellVariable)
        1
        >>> print vcvXvcv.rank
        1

    rank-1 `CellVariable` * rank-1 `FaceVariable`

        >>> vfvXvcv = vfv * vcv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> vcvXvfv = vcv * vfv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

    rank-1 `CellVariable` * Scalar

        >>> vcvXs = vcv * 3
        >>> print vcvXs
        [[0 3 6]
         [3 6 9]]
        >>> print isinstance(vcvXs, CellVariable)
        1
        >>> print vcvXs.rank
        1
        >>> sXvcv = 3 * vcv
        >>> print sXvcv
        [[0 3 6]
         [3 6 9]]
        >>> print isinstance(vcvXs, CellVariable)
        1
        >>> print vcvXs.rank
        1

    rank-1 `CellVariable` * Vector

        >>> vcvXv2 = vcv * (3,2)
        >>> print vcvXv2
        [[0 3 6]
         [2 4 6]]
        >>> vcvXv2 = vcv * ((3,),(2,))
        >>> print vcvXv2
        [[0 3 6]
         [2 4 6]]
        >>> print isinstance(vcvXv2, CellVariable)
        1
        >>> print vcvXv2.rank
        1
        >>> v2Xvcv = (3,2) * vcv
        >>> print v2Xvcv
        [[0 3 6]
         [2 4 6]]
        >>> v2Xvcv = ((3,),(2,)) * vcv
        >>> print v2Xvcv
        [[0 3 6]
         [2 4 6]]
        >>> print isinstance(v2Xvcv, CellVariable)
        1
        >>> print v2Xvcv.rank
        1
        
        >>> vcvXv3 = vcv * (3,2,1)
        >>> print vcvXv3
        [[0 2 2]
         [3 4 3]]
        >>> isinstance(vcvXv3, CellVariable)
        1
        >>> print vcvXv3.rank
        1
        >>> v3Xvcv = (3,2,1) * vcv 
        >>> print v3Xvcv
        [[0 2 2]
         [3 4 3]]
        >>> isinstance(v3Xvcv, CellVariable)
        1
        >>> print v3Xvcv.rank
        1

        >>> vcvXv4 = vcv * (3,2,1,0) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v4Xvcv = (3,2,1,0) * vcv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

    rank-1 `CellVariable` * `Variable` Scalar

        >>> vcvXsv = vcv * Variable(value=3)
        >>> print vcvXsv
        [[0 3 6]
         [3 6 9]]
        >>> print isinstance(vcvXsv, CellVariable)
        1
        >>> print vcvXsv.rank
        1
        >>> svXvcv = Variable(value=3) * vcv
        >>> print svXvcv
        [[0 3 6]
         [3 6 9]]
        >>> print isinstance(svXvcv, CellVariable)
        1
        >>> print svXvcv.rank
        1

    rank-1 `CellVariable` * `Variable` Vector
        
        >>> vcvXv2v = vcv * Variable(value=(3,2))
        >>> print vcvXv2v
        [[0 3 6]
         [2 4 6]]
        >>> vcvXv2v = vcv * Variable(value=((3,),(2,)))
        >>> print vcvXv2v
        [[0 3 6]
         [2 4 6]]
        >>> print isinstance(vcvXv2v, CellVariable)
        1
        >>> print vcvXv2v.rank
        1
        >>> v2vXvcv = Variable(value=(3,2)) * vcv
        >>> print v2vXvcv
        [[0 3 6]
         [2 4 6]]
        >>> v2vXvcv = Variable(value=((3,),(2,))) * vcv
        >>> print v2vXvcv
        [[0 3 6]
         [2 4 6]]
        >>> print isinstance(v2vXvcv, CellVariable)
        1
        >>> print v2vXvcv.rank
        1
        
        >>> vcvXv3v = vcv * Variable(value=(3,2,1))
        >>> print vcvXv3v
        [[0 2 2]
         [3 4 3]]
        >>> isinstance(vcvXv3v, CellVariable)
        1
        >>> print vcvXv3v.rank
        1
        >>> v3vXvcv = Variable(value=(3,2,1)) * vcv 
        >>> print v3vXvcv
        [[0 2 2]
         [3 4 3]]
        >>> isinstance(v3vXvcv, CellVariable)
        1
        >>> print v3vXvcv.rank
        1

        >>> vcvXv4v = vcv * Variable(value=(3,2,1,0)) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v4vXvcv = Variable(value=(3,2,1,0)) * vcv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

                    
    rank-1 `FaceVariable` * rank-1 FaceVariable

        >>> vfvXvfv = vfv * vfv
        >>> print vfvXvfv
        [[ 0  1  4  9  1  4  9 36  4  1]
         [ 1  4  9 16  9 16 25 81 36  9]]
        >>> isinstance(vfvXvfv, FaceVariable)
        1
        >>> print vfvXvfv.rank
        1

    rank-1 `FaceVariable` * Scalar

        >>> vfvXs = vfv * 3
        >>> print vfvXs
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 3  6  9 12  9 12 15 27 18  9]]
        >>> print isinstance(vfvXs, FaceVariable)
        1
        >>> print vfvXs.rank
        1
        >>> sXvfv = 3 * vfv
        >>> print sXvfv
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 3  6  9 12  9 12 15 27 18  9]]
        >>> print isinstance(sXvfv, FaceVariable)
        1
        >>> print sXvfv.rank
        1

    rank-1 `FaceVariable` * Vector

        >>> vfvXv2 = vfv * (3,2)
        >>> print vfvXv2
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 2  4  6  8  6  8 10 18 12  6]]
        >>> print isinstance(vfvXv2, FaceVariable)
        1
        >>> print vfvXv2.rank
        1
        >>> v2Xvfv = (3,2) * vfv
        >>> print v2Xvfv
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 2  4  6  8  6  8 10 18 12  6]]
        >>> print isinstance(v2Xvfv, FaceVariable)
        1
        >>> print v2Xvfv.rank
        1
        
        >>> vfvXv3 = vfv * (2,1,0) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v3Xvfv = (2,1,0) * vfv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int


        >>> vfvXv10 = vfv * (9,8,7,6,5,4,3,2,1,0)
        >>> print vfvXv10
        [[ 0  8 14 18  5  8  9 12  2  0]
         [ 9 16 21 24 15 16 15 18  6  0]]
        >>> isinstance(vfvXv10, FaceVariable)
        1
        >>> print vfvXv10.rank
        1
        >>> v10Xvfv = (9,8,7,6,5,4,3,2,1,0) * vfv
        >>> print v10Xvfv
        [[ 0  8 14 18  5  8  9 12  2  0]
         [ 9 16 21 24 15 16 15 18  6  0]]
        >>> isinstance(v10Xvfv, FaceVariable)
        1
        >>> print v10Xvfv.rank
        1

    rank-1 `FaceVariable` * `Variable` Scalar

        >>> vfvXsv = vfv * Variable(value=3)
        >>> print vfvXsv
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 3  6  9 12  9 12 15 27 18  9]]
        >>> print isinstance(vfvXsv, FaceVariable)
        1
        >>> print vfvXsv.rank
        1
        >>> svXvfv = Variable(value=3) * vfv
        >>> print svXvfv
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 3  6  9 12  9 12 15 27 18  9]]
        >>> print isinstance(svXvfv, FaceVariable)
        1
        >>> print svXvfv.rank
        1

    rank-1 `FaceVariable` * `Variable` Vector
        
        >>> vfvXv2v = vfv * Variable(value=(3,2))
        >>> print vfvXv2v
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 2  4  6  8  6  8 10 18 12  6]]
        >>> print isinstance(vfvXv2v, FaceVariable)
        1
        >>> print vfvXv2v.rank
        1
        >>> v2vXvfv = Variable(value=(3,2)) * vfv
        >>> print v2vXvfv
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 2  4  6  8  6  8 10 18 12  6]]
        >>> print isinstance(v2vXvfv, FaceVariable)
        1
        >>> print v2vXvfv.rank
        1
        
        >>> vfvXv3v = vfv * Variable(value=(2,1,0)) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v3vXvfv = Variable(value=(2,1,0)) * vfv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int


        >>> vfvXv10v = vfv * Variable(value=(9,8,7,6,5,4,3,2,1,0))
        >>> print vfvXv10v
        [[ 0  8 14 18  5  8  9 12  2  0]
         [ 9 16 21 24 15 16 15 18  6  0]]
        >>> isinstance(vfvXv10v, FaceVariable)
        1
        >>> print vfvXv10v.rank
        1
        >>> v10vXvfv = Variable(value=(9,8,7,6,5,4,3,2,1,0)) * vfv
        >>> print v10vXvfv
        [[ 0  8 14 18  5  8  9 12  2  0]
         [ 9 16 21 24 15 16 15 18  6  0]]
        >>> isinstance(v10vXvfv, FaceVariable)
        1
        >>> print v10vXvfv.rank
        1
        
        
    Scalar * `Variable` Scalar

        >>> sXsv = 3 * Variable(value=3)
        >>> print sXsv
        9
        >>> print isinstance(sXsv, Variable)
        1
        >>> svXs = Variable(value=3) * 3
        >>> print svXs
        9
        >>> print isinstance(svXs, Variable)
        1

    Scalar * `Variable` Vector
        
        >>> sXv2v = 3 * Variable(value=(3,2))
        >>> print sXv2v
        [9 6]
        >>> print isinstance(sXv2v, Variable)
        1
        >>> v2vXs = Variable(value=(3,2)) * 3
        >>> print v2vXs
        [9 6]
        >>> print isinstance(v2vXs, Variable)
        1
        
        
        
    Vector * `Variable` Scalar

        >>> vXsv = (3, 2) * Variable(value=3)
        >>> print vXsv
        [9 6]
        >>> print isinstance(vXsv, Variable)
        1
        >>> svXv = Variable(value=3) * (3, 2)
        >>> print svXv
        [9 6]
        >>> print isinstance(svXv, Variable)
        1

    Vector * `Variable` Vector
        
        >>> vXv2v = (3, 2) * Variable(value=(3,2))
        >>> print vXv2v
        [9 4]
        >>> print isinstance(vXv2v, Variable)
        1
        >>> v2vXv = Variable(value=(3,2)) * (3, 2)
        >>> print v2vXv
        [9 4]
        >>> print isinstance(v2vXv, Variable)
        1

        >>> vXv3v = (3, 2, 1) * Variable(value=(3,2)) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v3vXv = Variable(value=(3,2)) * (3, 2, 1) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        

    `Variable` Scalar * `Variable` Scalar

        >>> svXsv = Variable(value=3) * Variable(value=3)
        >>> print svXsv
        9
        >>> print isinstance(svXsv, Variable)
        1

    `Variable` Scalar * `Variable` Vector
        
        >>> svXv2v = Variable(value=3) * Variable(value=(3,2))
        >>> print svXv2v
        [9 6]
        >>> print isinstance(svXv2v, Variable)
        1
        >>> v2vXsv = Variable(value=(3,2)) * Variable(value=3)
        >>> print v2vXsv
        [9 6]
        >>> print isinstance(v2vXsv, Variable)
        1

        
    `Variable` Vector * `Variable` Vector
        
        >>> v2vXv2v = Variable(value=(3, 2)) * Variable(value=(3,2))
        >>> print v2vXv2v
        [9 4]
        >>> print isinstance(v2vXv2v, Variable)
        1
        
        >>> v3vXv2v = Variable(value=(3, 2, 1)) * Variable(value=(3,2)) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

    Test for weird bug that was appearing in inline. Caused by the intermediate
    operators not getting marked fresh.

        >>> class Alpha(Variable):
        ...     def __init__(self, var):
        ...         Variable.__init__(self)
        ...         self.var = self._requires(var)
        ...     def _calcValue(self):
        ...         return self.var.value

        >>> coeff = Variable()
        >>> alpha = Alpha(-coeff / 1)
        >>> print numerix.allclose(alpha.value, 0.0)
        True
        >>> coeff.value = (-10.0)
        >>> print numerix.allclose(alpha.value, 10)
        True
        >>> coeff.value = (10.0)
        >>> print numerix.allclose(alpha.value, -10)
        True

    Test to prevent divide by zero evaluation before value is
    requested.  The request is caused by the Variable requiring
    its unit to see whether it can do an inline calculation in
    `_UnaryOperatorVariable()`.
    
        >>> T = Variable()
        >>> from fipy import numerix
        >>> v = numerix.exp(-T / (1. *  T))

    Following is a test case for an error when turing a binOp into an array

        >>> print numerix.array(Variable(value=numerix.array([ 1.,])) * [ 1.,])
        [ 1.]

    It seems that numpy's __rmul__ coercion is very strange

        >>> type(numerix.array([1., 2.]) * Variable([1., 2.]))
        <class 'fipy.variables.binaryOperatorVariable.binOp'>

    Test inlining

        >>> v0 = Variable(numerix.ones(2, 'd'))
        >>> v1 = Variable(numerix.ones(2, 'd'))
        >>> v = v1 * v0
        >>> print v
        [ 1.  1.]
        >>> v0[1] = 0.5
        >>> print v
        [ 1.   0.5]
        
    Test inline indexing

        >>> mesh = Grid2D(nx=3, ny=3)
        >>> v1 = CellVariable(mesh=mesh, value=numerix.arange(9))
        >>> a = v1 * (1, -1)
        >>> print a
        [[ 0  1  2  3  4  5  6  7  8]
         [ 0 -1 -2 -3 -4 -5 -6 -7 -8]]
        >>> x, y = mesh.cellCenters
        >>> v1[(x == 0.5) & (y == 0.5)] = 0
        >>> print a
        [[ 0  1  2  3  4  5  6  7  8]
         [ 0 -1 -2 -3 -4 -5 -6 -7 -8]]

    """
    pass

def _test(): 
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()
    
if __name__ == "__main__": 
    _test()



#!/usr/bin/env python

## 
 # -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "mesh1D.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #  Author: James O'Beirne <james.obeirne@gmail.com>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution
 #  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

"""Generic mesh class using numerix to do the calculations

    Meshes contain cells, faces, and vertices.

    This is built for a non-mixed element mesh.
"""

from fipy.tools import numerix
from fipy.tools.numerix import MA
from fipy.tools import serialComm

from fipy.meshes.mesh import Mesh
from fipy.meshes.representations.meshRepresentation import _MeshRepresentation
from fipy.meshes.topologies.meshTopology import _Mesh1DTopology

__all__ = ["Mesh1D"]

class Mesh1D(Mesh):
    def __init__(self, vertexCoords, faceVertexIDs, cellFaceIDs, communicator=serialComm, _RepresentationClass=_MeshRepresentation, _TopologyClass=_Mesh1DTopology):
        super(Mesh1D, self).__init__(vertexCoords=vertexCoords, faceVertexIDs=faceVertexIDs, cellFaceIDs=cellFaceIDs, communicator=communicator, 
                                     _RepresentationClass=_RepresentationClass, _TopologyClass=_TopologyClass)

    def _calcScaleArea(self):
        return 1.

    def _calcScaleVolume(self):
        return self.scale['length']

    def _calcFaceAreas(self):
        return numerix.ones(self.numberOfFaces, 'd')

    def _calcFaceNormals(self):
        faceNormals = numerix.array((numerix.ones(self.numberOfFaces, 'd'),))
        # The left-most face has neighboring cells None and the left-most cell.
        # We must reverse the normal to make fluxes work correctly.
        if self.numberOfFaces > 0:
            faceNormals[...,0] = -faceNormals[...,0]
        return faceNormals

    def _calcFaceTangents(self):
        faceTangents1 = numerix.zeros(self.numberOfFaces, 'd')[numerix.NewAxis, ...]
        faceTangents2 = numerix.zeros(self.numberOfFaces, 'd')[numerix.NewAxis, ...]
        return faceTangents1, faceTangents2    

    def _translate(self, vector):
        newCoords = self.vertexCoords + vector
        newmesh = Mesh1D(newCoords, numerix.array(self.faceVertexIDs), numerix.array(self.cellFaceIDs))
        return newmesh

    def __mul__(self, factor):
        newCoords = self.vertexCoords * factor
        newmesh = Mesh1D(newCoords, numerix.array(self.faceVertexIDs), numerix.array(self.cellFaceIDs))
        return newmesh

    @property
    def _VTKCellType(self):
        try:
            from tvtk.api import tvtk
        except ImportError as e:
            from enthought.tvtk.api import tvtk
        return tvtk.Line().cell_type




#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "cellToFaceVariable.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution
 #  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

__all__ = []

from fipy.variables.faceVariable import FaceVariable
from fipy.tools import numerix

class _CellToFaceVariable(FaceVariable):
    def __init__(self, var):
        FaceVariable.__init__(self, mesh=var.mesh, elementshape=var.shape[:-1])
        self.var = self._requires(var)

    def _calcValue(self):
        alpha = self.mesh._faceToCellDistanceRatio
        id1, id2 = self.mesh._adjacentCellIDs
        
        return self._calcValue_(alpha=alpha, id1=id1, id2=id2)

    def release(self, constraint):
        """Remove `constraint` from `self`
        
        >>> from fipy import *
        >>> m = Grid1D(nx=3)
        >>> v = CellVariable(mesh=m, value=m.cellCenters[0])
        >>> c0 = Constraint(0., where=m.facesLeft)
        >>> v.constrain(c0)
        >>> c1 = Constraint(3., where=m.facesRight)
        >>> v.faceValue.constrain(c1)
        >>> print v.faceValue
        [ 0.  1.  2.  3.]
        >>> v.faceValue.release(constraint=c0)
        >>> print v.faceValue
        [ 0.5  1.   2.   3. ]
        >>> v.faceValue.release(constraint=c1)
        >>> print v.faceValue
        [ 0.5  1.   2.   2.5]
        """
        try:
            super(FaceVariable, self).constraints.remove(constraint)
        except ValueError:
            self.var.release(constraint=constraint)

    @property
    def constraints(self):
        if hasattr(self.var, "faceConstraints"):
            faceConstraints = self.var.faceConstraints
        else:
            faceConstraints = []
        
        return super(_CellToFaceVariable, self).constraints + faceConstraints

    @property
    def constraintMask(self):
        if hasattr(self, '_constraintMask'):
            if hasattr(self.var, 'faceConstraints'):
                for constraint in self.var.faceConstraints:
                    self._constraintMask._requires(constraint.where)

        return super(_CellToFaceVariable, self).constraintMask
            
    def __getstate__(self):
        return dict(var=self.var)
        
    def __setstate__(self, dict):
        self.__init__(**dict)

def _test(): 
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()
    
if __name__ == "__main__": 
    _test() 



__docformat__ = 'restructuredtext'

from fipy.variables.cellToFaceVariable import _CellToFaceVariable
from fipy.tools.numerix import exp, take, where

__all__ = ["ScharfetterGummelFaceVariable"]

class ScharfetterGummelFaceVariable(_CellToFaceVariable):
    def __init__(self, var, boundaryConditions=()):
        _CellToFaceVariable.__init__(self, var)
        self.bcs = boundaryConditions
        
    def _calcValue_(self, alpha, id1, id2):
        cell1 = take(self.var,id1)
        cell2 = take(self.var,id2)
        delta = cell1 - cell2
        
        eps = 1e-14
        value = where(abs(delta) < eps, 1. / exp(delta), 0.)
        delta = where(abs(delta) < eps, eps, delta) 
        value = where((abs(delta) > eps) & (delta < 100), 
                      delta / (exp(delta) - 1), value)
                            
        value *= exp(cell1) 
        
        for bc in self.bcs:
            if isinstance(bc, FixedValue):
                value[bc.faces.value] = bc._value
        
        return value



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "gist2DViewer.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution
 #  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

from fipy.tools import numerix
from fipy.viewers.gistViewer.gistViewer import _GistViewer
from fipy.tools.decorators import deprecateGist
__all__ = ["Gist2DViewer"]

@deprecateGist
class Gist2DViewer(_GistViewer):
    """Displays a contour plot of a 2D `CellVariable` object.
    """
    
    __doc__ += _GistViewer._test2D(viewer="Gist2DViewer")
    __doc__ += _GistViewer._test2Dirregular(viewer="Gist2DViewer")
    
    def __init__(self, vars, title=None, palette='heat.gp', grid=True, dpi=75, limits={}, **kwlimits):
        """Creates a `Gist2DViewer`.
        
        :Parameters:
          vars
            a `CellVariable` object.
          title
            displayed at the top of the `Viewer` window
          palette
            the color scheme to use for the image plot. Default is 
            `heat.gp`. Another choice would be `rainbow.gp`.
          grid
            whether to show the grid lines in the plot.
          limits : dict
            a (deprecated) alternative to limit keyword arguments
          xmin, xmax, ymin, ymax, datamin, datamax
            displayed range of data. Any limit set to 
            a (default) value of `None` will autoscale.
        """
        kwlimits.update(limits)
        _GistViewer.__init__(self, vars=vars, title=" ", dpi=dpi, **kwlimits)
                            
        self.palette = palette
        self.grid = grid
        
    def _getSuitableVars(self, vars):
        from fipy.variables.cellVariable import CellVariable
        vars = [var for var in _GistViewer._getSuitableVars(self, vars) \
          if (var.mesh.dim == 2 and isinstance(var, CellVariable))]
        if len(vars) == 0:
            from fipy.viewers import MeshDimensionError
            raise MeshDimensionError("Can only plot 2D data")
        # this viewer can only display one variable
        return [vars[0]]

    def _plot(self):
        import gist
    
        gist.window(self.id, wait = 1)
        gist.animate(1)
        gist.pltitle(self.title)
        gist.palette(self.palette)
        gist.gridxy(self.grid)

        if self.limits != None:
            gist.limits(self._getLimit('xmin'), 
                        self._getLimit('xmax'), 
                        self._getLimit('ymin'), 
                        self._getLimit('ymax'))

    def plot(self, filename = None):
        """
        Plot the `CellVariable` as a contour plot.
        """
        self._plot()

        datamin = self._getLimit(('datamin', 'zmin'))
        datamax = self._getLimit(('datamax', 'zmax'))
        
        if datamin == 'e':
            datamin = None
        
        if datamax == 'e':
            datamax = None
            
        datamin, datamax = self._autoscale(vars=self.vars,
                                           datamin=datamin,
                                           datamax=datamax)

        if datamax == datamin:
            datamax = datamin + 1e-10

        vertexIDs = self.mesh._orderedCellVertexIDs

        vertexCoords = self.mesh.vertexCoords

        xCoords = numerix.take(vertexCoords[0], vertexIDs).flatten("FORTRAN")
        yCoords = numerix.take(vertexCoords[1], vertexIDs).flatten("FORTRAN")

        import gist

        import Numeric
        gist.plfp(Numeric.array(numerix.array(self.vars[0])), yCoords, xCoords, self.mesh._numberOfFacesPerCell, cmin=datamin, cmax=datamax)

        import fipy.viewers.gistViewer.colorbar

        colorbar._color_bar(minz=datamin, maxz=datamax, ncol=240, zlabel=self.vars[0].name)

        _GistViewer.plot(self, filename = filename)

    def plotMesh(self, filename = None):
        self._plot()
        
        faceVertexIDs = self.mesh.faceVertexIDs
        vertexCoords = self.mesh.vertexCoords
        
        x0 = numerix.take(vertexCoords[0], faceVertexIDs[0])
        y0 = numerix.take(vertexCoords[1], faceVertexIDs[0])
        x1 = numerix.take(vertexCoords[0], faceVertexIDs[1])
        y1 = numerix.take(vertexCoords[1], faceVertexIDs[1])
        
        import gist
        
        gist.pldj(x0, y0, x1, y1)

        if filename is not None:
            import os.path
            root, ext = os.path.splitext(filename)
            if ext.lower() in (".eps", ".epsi"):
                gist.eps(root)
            else:
                gist.hcp_file(filename, dump = 1)
                gist.hcp()
                gist.hcp_finish(-1)

        gist.fma()

if __name__ == "__main__": 
    import fipy.tests.doctestPlus
    fipy.tests.doctestPlus.execButNoTest()



## -*-Pyth-*-
 # ###################################################################
 #  FiPy - a finite volume PDE solver in Python
 # 
 #  FILE: "gammaNoiseVariable.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This document was prepared at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this document is not subject to copyright
 # protection and is in the public domain.  gammaNoiseVariable.py
 # is an experimental work.  NIST assumes no responsibility whatsoever
 # for its use by other parties, and makes no guarantees, expressed
 # or implied, about its quality, reliability, or any other characteristic.
 # We would appreciate acknowledgement if the document is used.
 # 
 # This document can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution of
 #  this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

from fipy.tools.numerix import random
from fipy.variables.noiseVariable import NoiseVariable

__all__ = ["GammaNoiseVariable"]

class GammaNoiseVariable(NoiseVariable):
    r"""
    Represents a gamma distribution of random numbers with the probability
    distribution
    
    .. math::
    
       x^{\alpha - 1}\frac{\beta^\alpha e^{-\beta x}}{\Gamma(\alpha)}
       
    with a shape parameter :math:`\alpha`, a rate parameter :math:`\beta`, and 
    :math:`\Gamma(z) = \int_0^\infty t^{z - 1}e^{-t}\,dt`.

    Seed the random module for the sake of deterministic test results.

    >>> from fipy import numerix
    >>> numerix.random.seed(1)

    We generate noise on a uniform cartesian mesh
           
    >>> from fipy.variables.variable import Variable
    >>> alpha = Variable()
    >>> beta = Variable()
    >>> from fipy.meshes import Grid2D
    >>> noise = GammaNoiseVariable(mesh = Grid2D(nx = 100, ny = 100), shape = alpha, rate = beta)
           
    We histogram the root-volume-weighted noise distribution
    
    >>> from fipy.variables.histogramVariable import HistogramVariable
    >>> histogram = HistogramVariable(distribution = noise, dx = 0.1, nx = 300)
           
    and compare to a Gaussian distribution

    >>> from fipy.variables.cellVariable import CellVariable
    >>> x = CellVariable(mesh=histogram.mesh, value=histogram.mesh.cellCenters[0])
    >>> from scipy.special import gamma as Gamma # doctest: +SCIPY
    >>> from fipy.tools.numerix import exp
    >>> gammadist = (x**(alpha - 1) * (beta**alpha * exp(-beta * x)) / Gamma(alpha)) # doctest: +SCIPY    

    >>> if __name__ == '__main__':
    ...     from fipy import Viewer
    ...     viewer = Viewer(vars=noise, datamin=0, datamax=30)
    ...     histoplot = Viewer(vars=(histogram, gammadist), 
    ...                        datamin=0, datamax=1)
    
    >>> from fipy.tools.numerix import arange
    
    >>> for shape in arange(1,8,1):
    ...     alpha.value = shape
    ...     for rate in arange(0.5,2.5,0.5):
    ...         beta.value = rate
    ...         if __name__ == '__main__':
    ...             import sys
    ...             print >>sys.stderr, "alpha: %g, beta: %g" % (alpha, beta)
    ...             viewer.plot()
    ...             histoplot.plot()

    >>> print abs(noise.faceGrad.divergence.cellVolumeAverage) < 5e-15
    1

    .. image:: fipy/variables/gamma.*
      :scale: 25
      :align: center
      :alt: random values with a gamma distribution

    .. image:: fipy/variables/gamma-histogram.*
      :scale: 25
      :align: center
      :alt: histogram of random values with a gamma distribution

    """
    def __init__(self, mesh, shape, rate, name = '', hasOld = 0):
        r"""
        :Parameters:
            - `mesh`: The mesh on which to define the noise.
            - `shape`: The shape parameter, :math:`\alpha`.
            - `rate`: The rate or inverse scale parameter, :math:`\beta`.
                 
        """
        NoiseVariable.__init__(self, mesh = mesh, name = name, hasOld = hasOld)
        self.shapeParam = self._requires(shape)
        self.rate = self._requires(rate)
    
    def random(self):
        return random.gamma(shape=self.shapeParam, scale=self.rate, 
                            size=[self.mesh.globalNumberOfCells])

def _test(): 
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()
    
if __name__ == "__main__": 
    _test() 

